# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class ServiceInterface(object):
    def query(self, sql, limit=50000, offset=0, acceptPartial=False, **kwargs):
        raise NotImplementedError()

    def projects(self, **kwargs):
        raise NotImplementedError()

    def jobs(self, **kwargs):
        raise NotImplementedError()

    def resume_job(self, **kwargs):
        raise NotImplementedError()

    def tables_and_columns(self, **kwargs):
        raise NotImplementedError()

    def tables_in_hive(self, **kwargs):
        raise NotImplementedError()

    def cube_desc(self, name, **kwargs):
        raise NotImplementedError()

    def model_desc(self, name, **kwargs):
        raise NotImplementedError()

    def models(self, **kwargs):
        raise NotImplementedError()

    def cubes(self, **kwargs):
        raise NotImplementedError()

    def get_authentication(self, **kwargs):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import os

from sqlalchemy import create_engine, inspect

from kylinpy.utils.compat import as_unicode


class TestDialect(object):
    dsn = 'kylin://ADMIN:KYLIN@sandbox:7070/learn_kylin'
    cn_dsn = 'kylin://用户:密码@sandbox:7070/中文project'

    def json_loads(self, filename):
        here = os.path.abspath(os.path.dirname(__file__))
        f = open(os.path.join(here, 'fixtures', '{}.json'.format(filename)), 'r').read()
        return json.loads(f)

    @property
    def engine(self):
        return create_engine(self.dsn)

    def test_connection(self):
        engine = create_engine(self.cn_dsn)
        assert as_unicode(engine.url) == self.cn_dsn

        engine = create_engine('kylin://aa@bb123:aa@bb123@hello.world.com:1024/learn_kylin')
        assert engine.url.username == 'aa@bb123'
        assert engine.url.password == 'aa@bb123'
        assert engine.url.host == 'hello.world.com'
        assert engine.url.port == 1024
        assert engine.url.database == 'learn_kylin'

    def test_query(self, v1_api):
        engine = create_engine('kylin://ADMIN:KYLIN@sandbox/learn_kylin')
        rp = engine.execute('select count(*) from kylin_sales')
        assert [row[0] for row in rp.fetchall()] == [10000]

    def test_table_names(self, v1_api):
        engine = create_engine('kylin://ADMIN:KYLIN@sandbox/learn_kylin')
        assert engine.table_names() == [
            'KYLIN_ACCOUNT',
            'KYLIN_CAL_DT',
            'KYLIN_CATEGORY_GROUPINGS',
            'KYLIN_COUNTRY',
            'KYLIN_SALES',
        ]

    def test_get_schema_names(self, v1_api):
        engine = create_engine('kylin://ADMIN:KYLIN@sandbox/learn_kylin')
        insp = inspect(engine)
        assert insp.get_schema_names() == ['DEFAULT']

    def test_get_columns(self, v1_api):
        engine = create_engine('kylin://ADMIN:KYLIN@sandbox/learn_kylin')
        insp = inspect(engine)
        assert [c.get('name') for c in insp.get_columns('KYLIN_SALES', 'DEFAULT')] == [
            'TRANS_ID',
            'PART_DT',
            'LSTG_FORMAT_NAME',
            'LEAF_CATEG_ID',
            'LSTG_SITE_ID',
            'PRICE',
            'SELLER_ID',
            'BUYER_ID',
            'OPS_USER_ID',
            'OPS_REGION',
        ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime
import re

from kylinpy.exceptions import KylinUnsupportedType
from kylinpy.logger import logger
from kylinpy.utils.compat import text_type

true_pattern = re.compile(r'^true$', flags=re.IGNORECASE)

KylinType = dict(
    CHAR=text_type,
    VARCHAR=text_type,
    STRING=text_type,
    DECIMAL=float,
    DOUBLE=float,
    FLOAT=float,
    BIGINT=int,
    LONG=int,
    INTEGER=int,
    INT=int,
    TINYINT=int,
    SMALLINT=int,
    INT4=int,
    LONG8=int,
    BOOLEAN=lambda x: bool(re.search(true_pattern, x)),
    DATE=lambda x: datetime.strptime(x, '%Y-%m-%d').date(),
    DATETIME=lambda x: datetime.strptime(x, '%Y-%m-%d %H:%M:%S'),
    TIMESTAMP=lambda x: datetime.strptime(x.split('.')[0], '%Y-%m-%d %H:%M:%S'),
)


def _convert_type(_type):
    # remove parenthesis, such as 'DECIMAL(2, 5)'
    _type = re.sub(r'\(.*\)', '', _type)
    # remove some decorator in type string, such as 'BIGINT NOT NULL'
    _type = re.sub(r'\s.*', '', _type)
    return str(_type).upper()


def kylin_to_python(_type, s):
    try:
        return KylinType[_convert_type(_type)](s) if s else s
    except KeyError:
        logger.error(
            'CONVERT ERROR, type: {}, raw string: {}'.format(_type, s))
        raise KylinUnsupportedType(_type)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json as _json
import ssl

from kylinpy.utils.compat import urllib
from kylinpy.utils.compat import HTTPSHandler
from kylinpy.utils.compat import urlencode
from kylinpy.utils.compat import HTTPError
from kylinpy.client.exceptions import handle_error
from kylinpy.logger import logger


class Response(object):
    def __init__(self, response):
        self._status_code = response.getcode()
        self._body = response.read()
        self._headers = response.info()
        self._url = response.url

    @property
    def status_code(self):
        return self._status_code

    @property
    def url(self):
        return self._url

    @property
    def body(self):
        return self._body

    @property
    def headers(self):
        return self._headers

    def json(self):
        if self.body:
            return _json.loads(self.body.decode('utf-8'))
        else:
            return None


class Client(object):
    """Quickly and easily access any REST or REST-like API."""

    def __init__(
        self,
        host,
        request_headers=None,
        prefix=None,
        timeout=None,
        unverified=None,
        is_debug=False,
    ):
        self.host = host.rstrip('/')
        self.request_headers = request_headers or {}
        self.prefix = prefix
        self.timeout = timeout
        self.unverified = unverified
        self.is_debug = is_debug

    def _build_url(self, endpoint=None, params=None):
        if self.prefix:
            url = '{}/{}'.format(self.host, self.prefix.strip('/'))
        else:
            url = self.host

        if endpoint:
            url = '{}/{}'.format(url, endpoint.strip('/'))

        if params:
            params = {k: v for k, v in params.items() if v is not None}
            url_values = urlencode(sorted(params.items()), True)
            url = '{}?{}'.format(url, url_values)

        return url

    def _update_headers(self, request_headers):
        self.request_headers.update(request_headers)

    def _mask_auth_headers(self, headers):
        if self.is_debug:
            return headers

        _headers = {}
        for (key, value) in headers.items():
            if key.lower() == 'authorization':
                _headers[key] = '******'
            else:
                _headers[key] = value
        return _headers

    def _make_request(self, opener, request, timeout=None):
        timeout = timeout or self.timeout
        logger.debug("""
==========================[QUERY]===============================
method: {} 
url: {} 
headers: {} 
body: {}
==========================[QUERY]===============================""".format(  # noqa
            request.get_method(),
            request.get_full_url(),
            self._mask_auth_headers(dict(request.header_items())),
            request.data))
        try:
            return opener.open(request, timeout=timeout)
        except HTTPError as err:
            exc = handle_error(err)
            exc.__cause__ = None
            raise exc

    def _request(self, method, endpoint,
                 params=None, json=None, headers=None, timeout=None):
        method = method.upper()
        request_data = None

        if headers:
            self._update_headers(headers)
        if json:
            request_data = _json.dumps(json).encode('utf-8')

        if self.unverified:
            opener = urllib.build_opener(
                HTTPSHandler(context=ssl._create_unverified_context()),
            )
        else:
            opener = urllib.build_opener()
        request = urllib.Request(self._build_url(endpoint, params), data=request_data)
        if self.request_headers:
            for key, value in self.request_headers.items():
                request.add_header(key, value)
        if request_data and ('Content-Type' not in self.request_headers):
            request.add_header('Content-Type', 'application/json')

        request.get_method = lambda: method
        return Response(self._make_request(opener, request, timeout=timeout))

    def get(self, endpoint, params=None, **kwargs):
        return self._request('get', endpoint, params=params, **kwargs)

    def delete(self, endpoint, **kwargs):
        return self._request('delete', endpoint, **kwargs)

    def post(self, endpoint, json=None, **kwargs):
        return self._request('post', endpoint, json=json, **kwargs)

    def put(self, endpoint, json=None, **kwargs):
        return self._request('put', endpoint, json=json, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.client import HTTPError
from kylinpy.kylinpy import Kylin
from kylinpy.utils.compat import as_unicode
from kylinpy.utils.kylin_types import kylin_to_python


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self._arraysize = 1
        self.rowcount = -1
        self.results = []
        self.fetched_rows = 0
        self._column_metas = []

    def callproc(self):
        pass

    def close(self):
        pass

    @property
    def description(self):
        return tuple([
            as_unicode(c['label']),
            c['columnTypeName'].lower(),
            c['displaySize'],
            None,
            c['precision'],
            c['scale'],
            c['isNullable'],
        ] for c in self._column_metas)

    def execute(self, query, parameters={}):
        resp = self.connection.query(query, **parameters)

        self._column_metas = resp.get('columnMetas')
        self.results = [tuple([
            kylin_to_python(self.description[col][1], cell)
            for (col, cell) in enumerate(row)
        ]) for row in resp['results']]
        self.rowcount = len(self.results)
        self.fetched_rows = 0

    def executemany(self, query, seq_params=[]):
        results = []
        for param in seq_params:
            self.execute(query, param)
            results.extend(self.results)

        self.results = results
        self.rowcount = len(self.results)
        self.fetched_rows = 0

    def fetchone(self):
        if self.fetched_rows < self.rowcount:
            row = self.results[self.fetched_rows]
            self.fetched_rows += 1
            return row
        else:
            return None

    def fetchmany(self, size=None):
        fetched_rows = self.fetched_rows
        size = size or self.arraysize
        self.fetched_rows = fetched_rows + size
        return self.results[fetched_rows: self.fetched_rows]

    def fetchall(self):
        fetched_rows = self.fetched_rows
        self.fetched_rows = self.rowcount
        return self.results[fetched_rows:]

    def nextset(self):
        pass

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, array_size):
        self._arraysize = array_size

    def setinputsizes(self):
        pass

    def setoutputsize(self):
        pass


class Connection(Kylin):
    paramstyle = 'pyformat'
    threadsafety = 2
    apilevel = '2.0'
    Error = HTTPError

    def __init__(self, **kwargs):
        super(Connection, self).__init__(**kwargs)

    @classmethod
    def connect(cls, **kwargs):
        return cls(**kwargs)

    def close(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        pass

    def cursor(self):
        return Cursor(self)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ._source_interface import (
    DimensionInterface, MeasureInterface, SourceInterface,
)
from ..utils.sqla_types import kylin_to_sqla

try:
    from sqlalchemy import sql
except ImportError:
    pass


class CubeSource(SourceInterface):
    source_type = 'cube'

    def __init__(self, cube_desc, model_desc, tables_and_columns):
        self.cube_desc = cube_desc
        self.model_desc = model_desc
        self.tables_and_columns = tables_and_columns

    @property
    def name(self):
        return self.cube_name

    @property
    def model_name(self):
        return self.model_desc.get('name')

    @property
    def cube_name(self):
        return self.cube_desc.get('name')

    @property
    def fact_table(self):
        fullname = self.model_desc.get('fact_table')
        alias = fullname.split('.')[1]
        return _Table(fullname, alias)

    @property
    def _model_lookups(self):
        # lookup tables in model
        return tuple((_.get('alias'), _) for _ in self.model_desc.get('lookups'))

    @property
    def lookups(self):
        # lookup-tables and middle-join tables (snowflake schema)
        lookups_and_joins = {}

        # lookup-tables in cube
        lookups_in_cube = set([d.table.alias for d in self.dimensions]) | self.measure_tables

        # adding index for model lookups
        model_lookups_map = {}
        for (idx, l) in enumerate(self._model_lookups):
            (alias, lookup) = l
            lookup['__idx'] = idx
            model_lookups_map[alias] = lookup

        for tbl in lookups_in_cube:
            join_table = model_lookups_map.get(tbl)
            while join_table:
                if join_table.get('alias') not in lookups_and_joins:
                    lookups_and_joins[join_table.get('alias')] = join_table

                foreign_table = join_table.get('join').get('foreign_key')[0].split('.')[0]
                if foreign_table == self.fact_table.alias:
                    join_table = None
                else:
                    join_table = model_lookups_map.get(foreign_table)

        # sorted by model lookups
        rv = lookups_and_joins.values()
        return tuple(sorted(rv, key=lambda x: x['__idx']))

    @property
    def dimensions(self):
        _dimensions = []
        for dim in self.cube_desc.get('dimensions'):
            table_alias = dim.get('table')
            table = dict(self._model_lookups).get(table_alias)
            table = table.get('table') if table else self.fact_table.fullname
            table_clz = _Table(table, table_alias)

            column = dim['column'] if dim['derived'] is None else dim['derived'][0]
            column_alias = dim['name']
            tbl_map = self.tables_and_columns
            description = dict(tbl_map[table_clz.fullname].get('columns')).get(column)
            column_clz = _Column(column, column_alias, description)

            _dimensions.append(_CubeDimension(table_clz, column_clz))
        return _dimensions

    @property
    def measures(self):
        _measures = []
        for measure in self.cube_desc.get('measures'):
            _measures.append(_CubeMeasure(measure))
        return _measures

    @property
    def measure_tables(self):
        _tables = []
        for m in self.measures:
            if m.value_tables:
                _tables.extend(m.value_tables)
        return set(_tables)

    @property
    def last_modified(self):
        return self.cube_desc.get('last_modified')

    @property
    def identity(self):
        return self.cube_desc.get('uuid')

    def _get_table_clause(self, tbl_clz):
        table_clause = sql.table(tbl_clz.name)
        table_clause.schema = tbl_clz.scheme
        return sql.alias(table_clause, tbl_clz.alias)

    @property
    def from_clause(self):
        _from_clause = self._get_table_clause(self.fact_table)

        for lookup in self.lookups:
            _join_clause_and = []
            for (idx, pk) in enumerate(lookup['join']['primary_key']):
                fk = lookup['join']['foreign_key'][idx]
                fk_table, fk_column = fk.split('.')
                pk_table, pk_column = pk.split('.')
                fk_table_quoted = sql.quoted_name(fk_table, True)
                fk_column_quoted = sql.quoted_name(fk_column, True)
                pk_table_quoted = sql.quoted_name(pk_table, True)
                pk_column_quoted = sql.quoted_name(pk_column, True)

                pk_column = sql.column(fk_column_quoted,
                                       _selectable=sql.table(fk_table_quoted))
                fk_column = sql.column(pk_column_quoted,
                                       _selectable=sql.table(pk_table_quoted))
                _join_clause_and.append(pk_column == fk_column)

            _lookup = _Table(lookup.get('table'), lookup.get('alias'))
            _is_left_join = lookup['join']['type'].lower() == 'left'
            _from_clause = sql.join(
                left=_from_clause,
                right=self._get_table_clause(_lookup),
                onclause=sql.and_(*_join_clause_and),
                isouter=_is_left_join,
            )
        return _from_clause

    def __repr__(self):
        return ('<Cube Instance by '
                'model_name: {self.model_name}, '
                'cube_name: {self.cube_name}>').format(**locals())


class _CubeDimension(DimensionInterface):
    def __init__(self, table_clz, column_clz):
        self.table = table_clz
        self.column = column_clz

    @property
    def datatype(self):
        return self.column.datatype

    @property
    def name(self):
        return '{}.{}'.format(self.table.alias, self.column.name)

    @property
    def verbose(self):
        return self.column.alias

    def __repr__(self):
        return '<Dimension: {}.{}>'.format(self.table.alias, self.column.alias)


class _CubeMeasure(MeasureInterface):
    def __init__(self, description):
        self._description = description
        self._function = description.get('function')
        self._paramter_stack = []

    @property
    def name(self):
        return self._description.get('name')

    @property
    def verbose(self):
        return self._description.get('name')

    @property
    def measure_type(self):
        return self._function.get('expression')

    @property
    def expression(self):
        _params = self._get_parameter_values(self._function.get('parameter'))
        return self._get_aggregations_exp(self.measure_type, _params)

    @property
    def value_tables(self):
        _values_columns = self._get_parameter_values(
            self._function.get('parameter'), 'column')
        if _values_columns:
            _columns = _values_columns.split(', ')
            return set(c.split('.')[0] for c in _columns)
        return None

    def _get_parameter_values(self, paramter_obj, param_type=None):
        """ return a parameter string from cube metrics function object. """
        if paramter_obj.get('next_parameter'):
            self._get_parameter_values(paramter_obj.get('next_parameter'), param_type)
        else:
            if param_type is None:
                self._paramter_stack.append(paramter_obj.get('value'))
            else:
                if paramter_obj.get('type') == param_type:
                    self._paramter_stack.append(paramter_obj.get('value'))
        return ', '.join(self._paramter_stack)

    def _get_aggregations_exp(self, aggregations_key, column_value):
        """return aggregations expression with the column value"""
        metrics_expression = {
            'COUNT_DISTINCT': 'COUNT (DISTINCT {})'.format(column_value),
            'COUNT': 'COUNT ({})'.format(column_value),
            'SUM': 'SUM ({})'.format(column_value),
            'AVG': 'AVG ({})'.format(column_value),
            'MIN': 'MIN ({})'.format(column_value),
            'MAX': 'MAX ({})'.format(column_value),
        }
        return metrics_expression.get(aggregations_key)

    def __repr__(self):
        return '<Measure: {}>'.format(self.name)


class _Table(object):
    def __init__(self, fullname, alias=None):
        self.fullname = fullname
        self.alias = alias

    @property
    def scheme(self):
        return self.fullname.split('.')[0]

    @property
    def name(self):
        return self.fullname.split('.')[1]

    def __repr__(self):
        return ('<Table '
                'name: {self.fullname}, '
                'alias: {self.alias}>').format(**locals())


class _Column(object):
    def __init__(self, name, alias, description):
        self.name = name
        self.alias = alias
        self.description = description

    @property
    def datatype(self):
        return str(kylin_to_sqla(self.description.get('type_NAME')))

    def __repr__(self):
        return ('<Column '
                'name: {self.name}, '
                'alias: {self.alias}>').format(**locals())



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.client import InternalServerError, UnauthorizedError
from kylinpy.exceptions import KylinQueryError
from ._service_interface import ServiceInterface


class _Api(object):
    """
    For testing convenience
    """
    @staticmethod
    def query(client, endpoint, **kwargs):
        return client.post(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def tables_and_columns(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def projects(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def jobs(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def resume_job(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def tables(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def model_desc(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def models(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def authentication(client, endpoint, **kwargs):
        rv = client.get(endpoint=endpoint, **kwargs).json().get('data')
        if rv == {}:
            raise UnauthorizedError
        return rv


class KE4Service(ServiceInterface):
    api = _Api

    def __init__(self, client, project=None):
        self.client = client
        self.project = project

    def query(self, sql, limit=50000, offset=0, acceptPartial=False, **kwargs):
        json_data = {
            'acceptPartial': acceptPartial,
            'limit': limit,
            'offset': offset,
            'project': self.project,
            'sql': sql,
        }
        kwargs.setdefault('json', json_data)
        try:
            response = self.api.query(self.client, '/query', **kwargs)
        except InternalServerError as err:
            raise KylinQueryError(err)

        err_message = response.get('exceptionMessage')
        if err_message:
            raise KylinQueryError(err_message)

        return response

    def projects(self, **kwargs):
        params = {
            'page_offset': 0,
            'page_size': 1000,
        }
        kwargs.setdefault('params', params)
        _projects = self.api.projects(self.client, '/projects', **kwargs)
        return _projects.get('value')

    def jobs(self, **kwargs):
        params = {
            'page_offset': 0,
            'page_size': 20,
        }
        kwargs.setdefault('params', params)
        _jobs = self.api.jobs(self.client, '/jobs', **kwargs)
        return _jobs.get('value')

    def resume_job(self, job_id, **kwargs):
        params = {
            'jobId': job_id,
        }
        kwargs.setdefault('params', params)
        res = self.api.resume_job(self.client, '/jobs/{0}/resume'.format(job_id), **kwargs)
        return res

    def tables_and_columns(self, **kwargs):
        params = {
            'project': self.project,
        }
        kwargs.setdefault('params', params)
        resp = self.api.tables_and_columns(self.client, '/query/tables_and_columns', **kwargs)
        tbl_pair = tuple(('{}.{}'.format(tbl.get('table_SCHEM'), tbl.get('table_NAME')), tbl) for tbl in resp)
        for tbl in tbl_pair:
            tbl[1]['columns'] = [(col['column_NAME'], col) for col in tbl[1]['columns']]
        return dict(tbl_pair)

    def tables_in_hive(self, **kwargs):
        params = {
            'project': self.project,
            'ext': True,
        }
        kwargs.setdefault('params', params)
        tables = self.api.tables(self.client, '/tables', **kwargs)
        tables = tables.get('tables')

        __tables_in_hive = {}
        for tbl in tables:
            db = tbl['database']
            name = tbl['name']
            fullname = '{}.{}'.format(db, name)
            __tables_in_hive[fullname] = tbl

        return __tables_in_hive

    def model_desc(self, name, **kwargs):
        _url = {
            'project': self.project,
            'model_name': name,
        }
        _model_desc = self.api.model_desc(
            self.client,
            '/models/{project}/{model_name}/model_desc'.format(**_url), **kwargs)
        return _model_desc

    def models(self, **kwargs):
        params = {
            'project': self.project,
            'page_offset': 0,
            'page_size': 1000,
        }
        kwargs.setdefault('params', params)
        _models = self.api.models(self.client, '/models', **kwargs)
        return _models.get('value')

    def get_authentication(self, **kwargs):
        return self.api.authentication(self.client, '/user/authentication', **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import time

try:
    # Python 3
    import urllib.request as urllib  # noqa
    from urllib.request import urlparse  # noqa
    from urllib.request import HTTPSHandler  # noqa
    from urllib.parse import urlencode  # noqa
    from urllib.parse import parse_qsl  # noqa
    from urllib.error import HTTPError  # noqa
except ImportError:
    # Python 2
    import urllib2 as urllib  # noqa
    from urlparse import urlparse  # noqa
    from urlparse import parse_qsl  # noqa
    from urllib2 import HTTPError  # noqa
    from urllib2 import HTTPSHandler  # noqa
    from urllib import urlencode  # noqa

PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str
    integer_types = int
    class_types = type
    text_type = str
    binary_type = bytes

    def as_unicode(s):
        if isinstance(s, bytes):
            return s.decode('utf-8')
        return str(s)

else:
    import types
    string_types = basestring  # noqa
    integer_types = (int, long)  # noqa
    class_types = (type, types.ClassType)  # noqa
    text_type = unicode  # noqa
    binary_type = str

    def as_unicode(s):
        if isinstance(s, str):
            return s.decode('utf-8')
        return unicode(s)  # noqa


def to_seconds(date):
    """ return timestamp for date object compatible py2 """
    return time.mktime(date.timetuple())



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.utils.sqla_types import kylin_to_sqla


def test_string():
    string_obj = kylin_to_sqla('CHAR(64) hello this is long descrition')
    assert string_obj.length == 64

    string_obj = kylin_to_sqla('STRING(64) hello this is long descrition')
    assert string_obj.length == 64

    string_obj = kylin_to_sqla('VARCHAR(64) hello this is long descrition')
    assert string_obj.length == 64


def test_float():
    decimal_obj = kylin_to_sqla('DECIMAL(20,6)')
    assert decimal_obj.precision == 20
    assert decimal_obj.scale == 6

    decimal_obj = kylin_to_sqla('DOUBLE(20)')
    assert decimal_obj.precision == 20

    decimal_obj = kylin_to_sqla('FLOAT(20)')
    assert decimal_obj.precision == 20


def test_int():
    assert str(kylin_to_sqla('BIGINT')) == 'BIGINT'
    assert str(kylin_to_sqla('INTEGER')) == 'INTEGER'
    assert str(kylin_to_sqla('INT')) == 'INTEGER'
    assert str(kylin_to_sqla('TINYINT')) == 'SMALLINT'
    assert str(kylin_to_sqla('SMALLINT')) == 'SMALLINT'
    assert str(kylin_to_sqla('INT4')) == 'BIGINT'
    assert str(kylin_to_sqla('LONG8')) == 'BIGINT'


def test_others():
    assert str(kylin_to_sqla('BOOLEAN')) == 'BOOLEAN'
    assert str(kylin_to_sqla('DATE')) == 'DATE'
    assert str(kylin_to_sqla('DATETIME')) == 'DATETIME'
    assert str(kylin_to_sqla('TIMESTAMP')) == 'TIMESTAMP'



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.kylinpy import create_kylin


class TestTableSource:
    def test_table_source(self, v1_api):
        project = create_kylin('kylin://username:password@example/kylin_sales')
        table = project.get_table_source('KYLIN_SALES', 'DEFAULT')
        assert table.name == 'KYLIN_SALES'
        assert table.schema == 'DEFAULT'
        assert [c.name for c in table.columns] == [
            'TRANS_ID',
            'PART_DT',
            'LSTG_FORMAT_NAME',
            'LEAF_CATEG_ID',
            'LSTG_SITE_ID',
            'PRICE',
            'SELLER_ID',
            'BUYER_ID',
            'OPS_USER_ID',
            'OPS_REGION',
        ]

    def test_v2_table_source(self, v2_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v2')
        table = project.get_table_source('KYLIN_SALES', 'DEFAULT')
        assert table.name == 'KYLIN_SALES'
        assert table.schema == 'DEFAULT'
        assert [c.name for c in table.columns] == [
            'TRANS_ID',
            'PART_DT',
            'LSTG_FORMAT_NAME',
            'LEAF_CATEG_ID',
            'LSTG_SITE_ID',
            'PRICE',
            'ITEM_COUNT',
            'SELLER_ID',
            'BUYER_ID',
            'OPS_USER_ID',
            'OPS_REGION',
        ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import base64
import logging

from kylinpy.client import Client as HTTPClient
from kylinpy.service import KylinService, KE3Service, KE4Service
from kylinpy.datasource import TableSource, CubeSource, KE4ModelSource
from kylinpy.utils.compat import as_unicode, urlparse, parse_qsl

SERVICES = {
    'v1': KylinService,
    'v2': KE3Service,
    'v4': KE4Service,
}


class Kylin(object):
    def __init__(self, host, username=None, password=None, port=7070, project=None, **connect_args):
        if host.startswith(('http://', 'https://')):
            _, self.host = host.split('://')
        else:
            self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.is_ssl = connect_args.get('is_ssl', None)
        self.prefix = connect_args.get('prefix', '/kylin/api')
        self.timeout = connect_args.get('timeout', 30)
        self.unverified = bool(connect_args.get('unverified', True))
        self.version = connect_args.get('version', 'v1')
        self.is_pushdown = bool(connect_args.get('is_pushdown', False))
        self.is_debug = bool(connect_args.get('is_debug', False))
        self.scheme = 'https' if self.is_ssl else 'http'
        self.project = project
        if self.is_debug:
            logging.basicConfig(level=logging.DEBUG)

    @property
    def service(self):
        return SERVICES[self.version](self._get_client(), self.project)

    def _get_client(self):
        headers = {
            'User-Agent': 'Kylin Python Client',
        }

        if self.version == 'v2':
            headers.update({'Accept': 'application/vnd.apache.kylin-v2+json'})

        if self.version == 'v4':
            headers.update({'Accept': 'application/vnd.apache.kylin-v4-public+json'})

        if self.username and self.password:
            headers.update(self.basic_auth_dump(self.username, self.password))

        return HTTPClient(
            host='{self.scheme}://{self.host}:{self.port}'.format(**locals()),
            prefix=self.prefix,
            timeout=self.timeout,
            request_headers=headers,
            unverified=self.unverified,
            is_debug=self.is_debug,
        )

    def basic_auth_dump(self, username, password):
        _auth = as_unicode('{}:{}').format(username, password)
        _auth = base64.b64encode(_auth.encode('utf-8')).decode('ascii')
        return {'Authorization': 'Basic {}'.format(_auth)}

    def projects(self):
        return self.service.projects()

    def jobs(self, **kwargs):
        return self.service.jobs(**kwargs)

    def query(self, sql, **parameters):
        return self.service.query(sql, **parameters)

    def get_all_tables(self, schema=None):
        if self.is_pushdown:
            _full_names = sorted(list(self.service.tables_in_hive().keys()))
        else:
            _full_names = sorted(list(self.service.tables_and_columns().keys()))

        if schema:
            _full_names = [t for t in _full_names if t.split('.')[0] == schema]
        return [t.split('.')[1] for t in _full_names]

    def get_all_schemas(self):
        if self.is_pushdown:
            _full_names = sorted(list(self.service.tables_in_hive().keys()))
        else:
            _full_names = sorted(list(self.service.tables_and_columns().keys()))
        return list(set(t.split('.')[0] for t in _full_names))

    def get_table_source(self, name, schema=None):
        if '.' in name:
            schema, name = name.split('.', 1)
        fullname = '{}.{}'.format(schema, name)
        if self.is_pushdown:
            return TableSource(name, schema, self.service.tables_in_hive().get(fullname))
        else:
            return TableSource(name, schema, self.service.tables_and_columns().get(fullname))

    def get_cube_source(self, name):
        if self.version == 'v4':
            _params = {
                'project': self.project,
                'page_offset': 0,
                'page_size': 1000,
                'model_name': name,
                'exact': True,
            }
            model_desc = self.service.models(params=_params)[0]
            return KE4ModelSource(
                model_desc=model_desc,
                tables_and_columns=self.service.tables_and_columns(),
            )

        cube_desc = self.service.cube_desc(name)
        model_name = cube_desc.get('model_name')
        return CubeSource(
            cube_desc=cube_desc,
            model_desc=self.service.model_desc(model_name),
            tables_and_columns=self.service.tables_and_columns(),
        )

    def __str__(self):
        if self.project:
            dsn = ('{self.scheme}://'
                   '{self.username}:{self.password}@'
                   '{self.host}:{self.port}'
                   '/{self.project}')
        else:
            dsn = ('{self.scheme}://'
                   '{self.username}:{self.password}@'
                   '{self.host}:{self.port}')
        return dsn.format(**locals())

    def __repr__(self):
        return '<Kylinpy instance: {}>'.format(str(self))


def create_kylin(dsn):
    url = urlparse(dsn)
    project = url.path.lstrip('/')
    port = url.port or 7070
    query = dict(parse_qsl(url.query) or {})
    return Kylin(url.hostname, url.username, url.password, port, project, **query)


# For compatibility with older versions
KylinCluster = Kylin
dsn_proxy = create_kylin



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.exceptions import NoSuchTableError
from ._source_interface import ColumnInterface
from ._source_interface import SourceInterface


class TableSource(SourceInterface):
    source_type = 'table'

    def __init__(self, name, schema, table_desc):
        if not table_desc:
            raise NoSuchTableError
        self._name = name
        self._schema = schema
        self.table_desc = table_desc

    @property
    def name(self):
        return self._name

    @property
    def schema(self):
        return self._schema

    @property
    def columns_map(self):
        return self.table_desc.get('columns')

    @property
    def columns(self):
        return [_Column(col) for col in self.columns_map]

    def __repr__(self):
        return ('<Hive Instance '
                'table_name: {self.name}>').format(**locals())


class _Column(ColumnInterface):
    def __init__(self, description):
        self.description = description[1] if \
            isinstance(description, tuple) else description

    @property
    def name(self):
        return self.description.get('name') or self.description.get('column_NAME')

    @property
    def datatype(self):
        return self.description.get('datatype') or self.description.get('type_NAME')

    def __repr__(self):
        return ('<Column '
                'name: {self.name}>').format(**locals())



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os.path
import json

import pytest


def read(version, filename):
    here = os.path.abspath(os.path.dirname(__file__))
    return json.load(open(os.path.join(here, version, filename)))


@pytest.fixture
def v1_api(mocker):
    cube_desc = read('v1', 'cube_desc.json')
    cubes = read('v1', 'cubes.json')
    models = read('v1', 'models.json')
    projects = read('v1', 'projects.json')
    jobs = read('v1', 'jobs.json')
    resume_job = read('v1', 'resume_job.json')
    query = read('v1', 'query.json')
    tables = read('v1', 'tables.json')
    tables_and_columns = read('v1', 'tables_and_columns.json')
    authentication = read('v1', 'authentication.json')

    mocker.patch('kylinpy.service.KylinService.api.cube_desc', return_value=cube_desc)
    mocker.patch('kylinpy.service.KylinService.api.cubes', return_value=cubes)
    mocker.patch('kylinpy.service.KylinService.api.models', return_value=models)
    mocker.patch('kylinpy.service.KylinService.api.projects', return_value=projects)
    mocker.patch('kylinpy.service.KylinService.api.jobs', return_value=jobs)
    mocker.patch('kylinpy.service.KylinService.api.resume_job', return_value=resume_job)
    mocker.patch('kylinpy.service.KylinService.api.query', return_value=query)
    mocker.patch('kylinpy.service.KylinService.api.tables', return_value=tables)
    mocker.patch('kylinpy.service.KylinService.api.tables_and_columns', return_value=tables_and_columns)
    mocker.patch('kylinpy.service.KylinService.api.authentication', return_value=authentication)
    yield mocker


@pytest.fixture
def v2_api(mocker):
    cube_desc = read('v2', 'cube_desc.json').get('data')
    cubes = read('v2', 'cubes.json').get('data')
    models = read('v2', 'models.json').get('data')
    projects = read('v2', 'projects.json').get('data')
    jobs = read('v2', 'jobs.json').get('data')
    resume_job = read('v2', 'resume_job.json').get('data')
    query = read('v2', 'query.json').get('data')
    tables = read('v2', 'tables.json').get('data')
    tables_and_columns = read('v2', 'tables_and_columns.json').get('data')
    authentication = read('v2', 'authentication.json').get('data')

    mocker.patch('kylinpy.service.KE3Service.api.cube_desc', return_value=cube_desc)
    mocker.patch('kylinpy.service.KE3Service.api.cubes', return_value=cubes)
    mocker.patch('kylinpy.service.KE3Service.api.models', return_value=models)
    mocker.patch('kylinpy.service.KE3Service.api.projects', return_value=projects)
    mocker.patch('kylinpy.service.KE3Service.api.jobs', return_value=jobs)
    mocker.patch('kylinpy.service.KE3Service.api.resume_job', return_value=resume_job)
    mocker.patch('kylinpy.service.KE3Service.api.query', return_value=query)
    mocker.patch('kylinpy.service.KE3Service.api.tables', return_value=tables)
    mocker.patch('kylinpy.service.KE3Service.api.tables_and_columns', return_value=tables_and_columns)
    mocker.patch('kylinpy.service.KE3Service.api.authentication', return_value=authentication)
    yield mocker


@pytest.fixture
def v4_api(mocker):
    projects = read('v4', 'projects.json').get('data')
    jobs = read('v4', 'jobs.json').get('data')
    resume_job = read('v4', 'resume_job.json').get('data')
    query = read('v4', 'query.json').get('data')
    tables = read('v4', 'tables.json').get('data')
    tables_and_columns = read('v4', 'tables_and_columns.json').get('data')
    authentication = read('v4', 'authentication.json').get('data')
    models = read('v4', 'models.json').get('data')
    model_desc = read('v4', 'model_desc.json').get('data')

    mocker.patch('kylinpy.service.KE4Service.api.projects', return_value=projects)
    mocker.patch('kylinpy.service.KE4Service.api.jobs', return_value=jobs)
    mocker.patch('kylinpy.service.KE4Service.api.resume_job', return_value=resume_job)
    mocker.patch('kylinpy.service.KE4Service.api.query', return_value=query)
    mocker.patch('kylinpy.service.KE4Service.api.tables', return_value=tables)
    mocker.patch('kylinpy.service.KE4Service.api.tables_and_columns', return_value=tables_and_columns)
    mocker.patch('kylinpy.service.KE4Service.api.authentication', return_value=authentication)
    mocker.patch('kylinpy.service.KE4Service.api.models', return_value=models)
    mocker.patch('kylinpy.service.KE4Service.api.model_desc', return_value=model_desc)
    yield mocker



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class KylinError(Exception):
    pass


class KylinUnauthorizedError(KylinError):
    pass


class KylinUserDisabled(KylinError):
    pass


class NoSuchTableError(KylinError):
    pass


class KylinQueryError(KylinError):
    pass


class KylinUnsupportedType(KylinError):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import codecs
import os.path
import re

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


setup(
    name='kylinpy',
    version=find_version('kylinpy', '__init__.py'),
    author='Yongjie Zhao',
    author_email='yongjie.zhao@kyligence.io',
    maintainer='Yongjie Zhao',
    maintainer_email='yongjie.zhao@kyligence.io',
    packages=find_packages(),
    url='https://github.com/Kyligence/kylinpy',
    license='MIT License',
    description='Apache Kylin Python Client Library',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    install_requires=[],
    extras_require={
        'sqlalchemy': ['sqlalchemy>=1.1.0'],
    },
    keywords=['apache kylin', 'kylin', 'kap', 'kyligence',
              'kyligence enterprise', 'cli', 'sqlalchemy dialect'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    entry_points={
        'sqlalchemy.dialects': ['kylin=kylinpy.sqla_dialect:KylinDialect'],
    },
)
